/*
 * FILE --- ga.hh
 * 
 * PURPOSE --- Header file for a utility that implements a Genetic 
 * Algorithm (GA) that is a stochastic direct search or optimization method.
 * It manipulates a string of numbers (a chromosome) in a manner similar to 
 * the laws of evolution and biology.
 *
 */

// Prevent Multiple Inclusion
#ifndef GA_HH
#define GA_HH

// Include Files
#include <stdio.h>
#include <math.h>

#include "fitness_param.hh"

// Local Definitions
#define GA_NO_ERROR 0
#define GA_MEM_ERROR 1
#define GA_MEM_NEGATIVE_FITNESS 2

#define ZERO  0.00000000001  // Compensate for roundoff error

// Chromosome is a string of genes that will be contained in each 
// individual
class CHROMOSOME
{
public:

  // Constructor
  CHROMOSOME(int _chrom_length);

  // Destructor
  ~CHROMOSOME();

  // Create a new cromosome with genes b/n the specified boundaries
  // The arrays _gene_low and _gene_high should be of length chrom_length
  void create_chrom(int *_gene_low, int *_gene_high);

  // Mutate the specified gene within the specified boundaries
  void mutate(int _gene, int _low, int _high);

  // Set the chromosome to the specified values 
  // The array _genes should be of length chrom_length
  void set_chrom(int *_genes);

  int chrom_length; // Length of the chromosome
  int *gene;        // String of genes (the chromosome) 

  int chrom_err;    // Variable to hold the error condition
};
typedef class CHROMOSOME Chromosome;

// Traits are decimal numbers that are decoded from the sections
// of each chromosome
class TRAITS
{
public:

  // Constructor
  TRAITS(int _num_traits);

  // Destructor
  ~TRAITS();

  int num_traits;    // Number of the traits in a chromosome
  double *trait_val; // The array of the trait values
  int trait_err;     // Variable to hold the error condition
};
typedef class TRAITS Traits;


// An individual is a member of a population. Each individual is 
// described by its chromosome. The chromosome determines an 
// individual's traits.  The individual's traits determine its fitness.
// The GA performs a search for the individual with the highest 
// possible fitness.
class INDIVIDUAL: public CHROMOSOME, public TRAITS
{
public:

  // Constructor
  INDIVIDUAL(int _chrom_lenght, int _num_traits); 

  // Destructor
  ~INDIVIDUAL();

  // Find the fitness of the individual
  // This function is to be modified by the user to provide
  // the application specific fitness function
  //  void get_fitness(Fitness_param *_fitness_param);
  //  void (*get_fitness) (Fitness_param *_fitness_param);

  double fitness; // Variable to hold the fitness of the individual
  int parents[2]; // the numbers of the parents of the individual

  int indv_err;  // Variable to hold the error condition
};
typedef class INDIVIDUAL Individual;

// A survivor is an individual who is selected to reproduce into the next 
// generation. In other words, its genetic material (chromosome) will 
// be passed to the the next generations.
class SURVIVOR: public CHROMOSOME
{
public:

  // Constructor
  SURVIVOR(int _chrom_length);

  // Destructor
  ~SURVIVOR();

  int surv_num; // The number of the survivor within the current population 
  int surv_err;  // Variable to hold the error condition
};
typedef class SURVIVOR Survivor;

// A population is a collection of individuals.
class POPULATION
{
public:

  // Constructor
  POPULATION(int _pop_size, int _chrom_lenght, int _num_traits);

  // Destructor
  ~POPULATION();

  // Set the boundaries of the traits within the chromosomes
  void set_trait_start(int *_trait_start);  

  // Set the trait offsets to be used in decoding the chromosomes
  void set_trait_offset(double *_trait_offset);

  // Set the valid ranges of the traits 
  void set_trait_low(double *_trait_low);
  void set_trait_high(double *_trait_high);

  // Set the place of the decimal point to be used in decoding the chromosomes
  void set_decimal(int *_decimal);

  // Set the valid ranges of the genes for the chromosoms
  void set_gene_low(int *_gene_low);
  void set_gene_high(int *_gene_high);

  // Set the crossover type 1 or 0
  // 0 = swap each gene with 50% probability
  // 1 = swap all the genes before a random selected site
  void set_cross_type(int _cross_type);

  // Set the crossover and mutation probabilities
  void set_cross_prob(double _cross_prob);
  void set_mutat_prob(double _mutat_prob);

  // Use elitism or not 1=on 0=off
  void set_elit_type(int _elit_type);

  // Use sign in decoding traits or not 1=signed 0=unsigned
  // If signed traits are used the first gene of each chromosome
  // determines the sign of the trait.
  // If this gene is 0-4, the trait is negative.
  //If this gene is 5-9, the trait is positive.
  void set_decode_type(int _decode_type);

  // Decode the chromosome of the individual into traits
  void decode_chrom(Individual *_individual);

  // Check if the chromosome of the individual is valid 
  int check_chrom(Individual *_individual);
  
  // Fix the chromosome of the individual (make it valid)
  // _flag specifies the trait to be fixed
  void fix_chrom(Individual *individual, int _flag);

  // Encode the traits of the individual into chromosome
  void encode_traits(Individual *_individual);

  // Find the fitness of all the individuals in the population
  void find_fitness(Fitness_param *_fitness_param); 

  // Create a random first generation
  void first_generation(void);  

  // Given a generation, create the next generation
  void next_generation(void);

  // Select the survivors from the given generation
  // The survivors are the individuals which will be reproduced
  // to form the next generation (mate and make offspring)
  void select_survivors(void);

  // Make the offspring from the survivors
  void make_offspring(void);

  // Mate (make crossover) _individual1 with _individual2
  void crossover(Individual *_individual1, Individual *_individual2);

  // Mutate the chromosome of the given individual
  void mutate(Individual *_individual);

  // Save statistical data to the specified file
  void statistics(const char *_stat_file);

  // Save the statistics about the best individuals the the specified file
  void best_statistics(const char *_best_stat_file);
  
  // Save the given generation to the specified file
  void save_generation(const char *_generation_file);

  Individual **members; // Array of the individuals in the population
  Survivor **survivors; // Array for the individuals that survive

  int *trait_start;  // The starting points (boundaries) for the traits 
  double *trait_offset; // The offsets for the traits
  double *trait_low;    // The lower ranges for the traits
  double *trait_high;   // The upper ranges for the traits
  int *decimal;         // Place of the decimal point in decoding traits
  int *gene_low;        // The lower ranges for the genes
  int *gene_high;       // The upper ranges for the genes

  double sumfitness;    // The sum of the fitness of all the individuals
  int bestmember;       // The number (index) of the best member
  int worstmember;      // The number (index) of the worst member

  int generation;       // The nember of the current generation  
  int crossovers;  // The nember of corssovers that occured in this generation
  int mutations;   // The nember of mutations that occured in this generation
  int pop_err;  // Variable to hold the error condition

private:

  int pop_size;  // Size of the population (constant for all generations)
  int chrom_length; // the chromosome lenght of all the individuals
  int num_traits;   // the number of the traits for all the individuals

  int use_sign;      // Use signed traits or not
  int use_elitism;   // Use elitism or not
  double cross_prob; // Crosover probability
  double mutat_prob; // Mutation probability
  int cross_type;    // Crossover type

};
typedef class POPULATION Population;


// Fitness function defined somewhere else 
void get_fitness(Individual *_individual, Fitness_param *_fitness_param);

#endif // GA_HH
